#include "dictionary.h"

Dictionary::Dictionary(const vector<string>& dataset, const int& lBound, const int& uBound, const int& updl, const int& delt, const int& eps, const int& maxReg, const double& regLen, const vector<bool>& passFlags, vector<double>& mmlMat, vector<double>& condMat, const string& logfile) {
    assert(dataset.size() > 0);
    assert(passFlags.size() >= 4);
    assert(mmlMat.size() == residueSize && condMat.size() == residueSize*residueSize);

    dataSet = dataset;
    mmlMatrix = mmlMat;
    condMatrix = condMat;
	lowerBound = lBound;
    upperBound = uBound;
    updateLen = updl;
    delta = delt;
    epsln = eps;
    maxLen = maxReg;
    regionLen = regLen;
    disectLegend = passFlags[0];
    pertLegend = passFlags[1];
    disctBreakdown = passFlags[2];
    prntCost = passFlags[3];
    logF = logfile;

    vector<Cost> nullConcept;
    for(size_t q = 0; q < dataSet.size(); q++)
        nullConcept.push_back(Cost(dataSet[q].size(), 0, delta));

    #pragma omp parallel for 
    for(size_t p = 0; p < dataSet.size(); p++){
        int seql = dataSet[p].size();
        for(int i = 1; i <= seql; i++){
            nullConcept[p](i-1,i,0) = mmlEstimate(dataSet[p].substr(i-1, 1), mmlMatrix);
            for (int j = i-2; j >= 0; j--){
                nullConcept[p](j,i,0) = nullConcept[p](j,i-1,0) + nullConcept[p](i-1,i,0);
            }
        }
    }
    double nullHypTotal = 0;
    #pragma omp parallel for reduction(+:nullHypTotal)
    for(size_t p = 0; p < dataSet.size(); p++){
        nullHypTotal += encodeWTNits(dataSet[p].size());
        nullHypTotal += nullConcept[p](0, dataSet[p].size(), 0);
    }

    nullML = encodeWTNits(dataSet.size()) + nullHypTotal;
    Concept nullCCPT;
    nullCCPT.costMat = nullConcept;
    nullCCPT._ind = -1;
    nullCCPT._pos = -1;
    nullCCPT._len = -1;
    nullCCPT.conceptML = 0;
    nullCCPT.usage.push_back(make_tuple(-1,-1,-1));
    conceptList.push_back(nullCCPT);

    for(int i = 0; i < maxLen/2; i++)
        conceptLen.push_back(vector<list<Concept>::iterator>(1, conceptList.begin()));

    nSize++;
    nSizeTrue++;
}

void Dictionary::_pop(){
	assert(nSize > 1 && conceptList.size() == nSize);
    dictML -= conceptList.back().conceptML;
    conceptLen[conceptList.back()._len].pop_back();
    conceptList.pop_back();
    
    nSize--;
    nSizeTrue--;
}

list<Concept>::iterator Dictionary::_randConcept(){
    assert(nSize > 1);
    int indx = rand() % (nSize-1) + 1;
    list<Concept>::iterator listIter = conceptList.begin();
    advance(listIter, indx);
    return listIter;
}

void Dictionary::_suppressConcept(const list<Concept>::iterator& listIter){
    assert(listIter != conceptList.end());
    
    vector<list<Concept>::iterator>::iterator itCCPTLen = find(conceptLen[(*listIter)._len].begin()+1, conceptLen[(*listIter)._len].end(), listIter);
    assert(itCCPTLen != conceptLen[(*listIter)._len].end());
    conceptLen[(*listIter)._len].erase(itCCPTLen); 

    nSizeTrue--;
    dictML -= (*listIter).conceptML;
    
    if(pertLegend){
        string pertLeg = "Del : (" + to_string((*listIter)._ind) + "," + to_string((*listIter)._pos) + "," + to_string((*listIter)._len) + ")\n";
        file_append(logF, pertLeg);
        cout << pertLeg;
    }    
}

void Dictionary::_unsuppressConcept(const list<Concept>::iterator& listIter){
    assert(listIter != conceptList.end());

    conceptLen[(*listIter)._len].push_back(listIter); 
    nSizeTrue++;
    dictML += (*listIter).conceptML;
}

void Dictionary::_newConcept(int& ind, int& pos, int& len){
    assert(dataSet.size() > 0);
    ind = rand() % dataSet.size();
    assert(ind >= 0 && ind < dataSet.size());
    
    assert(dataSet[ind].size() > lowerBound);
    pos = rand() % (dataSet[ind].size()-lowerBound);
    assert(pos >= 0 && pos <= (dataSet[ind].size()-lowerBound));

    len = lowerBound;
    if (dataSet[ind].size()-pos > lowerBound) {
        if(dataSet[ind].size()-pos >= upperBound) len = rand() % (upperBound-lowerBound) + lowerBound;
        else len = rand() % (dataSet[ind].size()-pos-lowerBound) + lowerBound;
    }
    assert(len >= lowerBound && len <= (dataSet[ind].size()-pos) && len <= upperBound); 
}

void Dictionary::_add(int& ind, int& pos, int& len){
    assert(ind >= 0 && ind < dataSet.size());
    assert(pos >= 0 && pos <= (dataSet[ind].size()-lowerBound));
    assert(len >= lowerBound && len <= (dataSet[ind].size()-pos)); 

    Concept ccpt;
    for(size_t q = 0; q < dataSet.size(); q++)
        ccpt.costMat.push_back(Cost (dataSet[q].size(), len, delta)); 

    string ccptSeq = dataSet[ind].substr(pos, len);
    populateCostMat(ccpt.costMat, ccptSeq);
    
    double ccptML = mmlEstimate(ccptSeq, mmlMatrix) + encodeWTNits(len);
    
    ccpt._ind = ind;
    ccpt._pos = pos;
    ccpt._len = len;
    ccpt.conceptML = ccptML;
    ccpt.usage.push_back(make_tuple(-1,-1,-1));
    conceptList.push_back(ccpt);

    list<Concept>::iterator listIter = conceptList.end();
    conceptLen[len].push_back(--listIter);
      
    nSize++;
    nSizeTrue++;
    dictML += ccptML;
    
    if(pertLegend){
        string pertLeg = "Add : (" + to_string(ind) + "," + to_string(pos) + "," + to_string(len) + ")\n";
        file_append(logF, pertLeg);
        cout << pertLeg;
    }    
}

void Dictionary::_remove(const list<Concept>::iterator& listIter){
    assert(listIter != conceptList.end());
    conceptList.erase(listIter);
    nSize--; 
}

void Dictionary::updateConceptLength(const int& ind, int& pos, int& len){
    assert(ind >= 0 && ind < dataSet.size()); 
    int row_size = dataSet[ind].size();
    assert(pos >= 0 && pos <= (row_size-lowerBound));   
    assert(len >= lowerBound && len <= (row_size-pos));
    assert(!(len == row_size-pos && len == lowerBound && pos == 0));

    int right, add;
    if(len == row_size-pos) add = 0;     //pos+1, len--
    else if(len == lowerBound) add = 1;      //pos-1, len++
    else add = rand() % 2;

    right = rand() % 2;
    if(pos == 0 && right == 0 && len == lowerBound) right = 1;    
    else if(pos == 0 && right == 0) add = 0;
    if(len == row_size-pos && len == lowerBound){
        add = 1;
        right = 0;
    } 

    if(right == 1) {
        int randRange = 1;
        if(add == 1) {
            if(updateLen > row_size-pos-len && row_size-pos-len > 1) randRange = row_size-pos-len-1;
            else if(updateLen <= row_size-pos-len && updateLen > 1) randRange = updateLen-1;
            else randRange = 1;
            assert(randRange > 0);
            int randVal = rand() % randRange + 1;
            len += randVal;
        } else {
            if(len-lowerBound >= updateLen && updateLen > 1) randRange = updateLen-1;
            else if(len-lowerBound < updateLen && len > lowerBound) randRange = len-lowerBound;
            else randRange = 1;
            assert(randRange > 0);
            int randVal = rand() % randRange + 1;
            len -= randVal;
        }
    } else {
        int randRange = 1;
        if(add == 1) {
            if(updateLen > row_size-pos-len && pos >= row_size-pos-len && row_size-pos-len > 1) randRange = row_size-pos-len-1;
            else if(updateLen > pos && row_size-pos-len >= pos && pos > 1) randRange = pos-1;
            else if(updateLen > 1 && updateLen <= row_size-pos-len && updateLen <= pos) randRange = updateLen-1; 
            else randRange = 1;
            assert(randRange > 0);
            int randVal = rand() % randRange + 1;
            len += randVal;
            pos -= randVal;
        } else {
            if(len-lowerBound >= updateLen && updateLen > 1) randRange = updateLen-1;
            else if(len-lowerBound < updateLen && len > lowerBound) randRange = len-lowerBound;
            else randRange = 1;
            assert(randRange > 0);
            int randVal = rand() % randRange + 1;
            len -= randVal;
            pos += randVal;
        }  
    }
    assert(ind >= 0 && ind < dataSet.size()); 
    assert(pos >= 0 && pos <= (row_size-lowerBound));   
    assert(len >= lowerBound && len <= (row_size-pos));
}

void Dictionary::_update(list<Concept>::iterator& listIter){
    assert(listIter != conceptList.end());
    
    int ind = (*listIter)._ind;
    int pos = (*listIter)._pos;
    int len = (*listIter)._len;
    tempCCPT = *listIter;

    updateConceptLength(ind, pos, len);
    Concept ccpt;
    ccpt._ind = ind;
    ccpt._pos = pos;
    ccpt._len = len;
    ccpt.usage.push_back(make_tuple(-1,-1,-1));

	for(size_t q = 0; q < dataSet.size(); q++)
        ccpt.costMat.push_back(Cost (dataSet[q].size(), len, delta));   

    string ccptSeq = dataSet[ind].substr(pos, len);
    populateCostMat(ccpt.costMat, ccptSeq);
    
    double ccptML = mmlEstimate(ccptSeq, mmlMatrix) + encodeWTNits(len);
    dictML -= (*listIter).conceptML;
    ccpt.conceptML = ccptML;
    dictML += ccptML;

    vector<list<Concept>::iterator>::iterator itCCPTLen = find(conceptLen[(*listIter)._len].begin()+1, conceptLen[(*listIter)._len].end(), listIter);
    assert(itCCPTLen != conceptLen[(*listIter)._len].end());
    conceptLen[(*listIter)._len].erase(itCCPTLen); 
    
    listIter = conceptList.erase(listIter);
    conceptList.insert(listIter, ccpt);
    conceptLen[len].push_back(--listIter);

    if(pertLegend){
        string pertLeg = "Upd : (" + to_string(ind) + "," + to_string(pos) + "," + to_string(len) + ")\n";
        file_append(logF, pertLeg);
        cout << pertLeg;
    }      
}

bool Dictionary::_swap(list<Concept>::iterator& listIter){
    assert(listIter != conceptList.end());
    
    int usgSize = (*listIter).usage.size();
    if(usgSize == 1){
        _suppressConcept(listIter);
        _remove(listIter);
        return false;
    }
    
    tempCCPT = *listIter;

    sort((*listIter).usage.begin(), (*listIter).usage.end());

    int ind = 0, pos = 0, len = 0;
    int r = rand() % (usgSize-1) + 1;
    assert(r > 0 && r < usgSize);
    ind = get<0>((*listIter).usage[r]);
    pos = get<1>((*listIter).usage[r]);
    len = get<2>((*listIter).usage[r]);
    
    assert(ind >= 0 && ind < dataSet.size());
    assert(pos >= 0 && pos <= (dataSet[ind].size()-lowerBound));
    assert(len >= lowerBound && len <= (dataSet[ind].size()-pos)); 

    Concept ccpt;
    ccpt._ind = ind;
    ccpt._pos = pos;
    ccpt._len = len;
    ccpt.usage.push_back(make_tuple(-1,-1,-1));

	for(size_t q = 0; q < dataSet.size(); q++)
        ccpt.costMat.push_back(Cost (dataSet[q].size(), len, delta));        
	
    string ccptSeq = dataSet[ind].substr(pos, len);
    populateCostMat(ccpt.costMat, ccptSeq);
   
    double ccptML = mmlEstimate(ccptSeq, mmlMatrix) + encodeWTNits(len);
    dictML -= (*listIter).conceptML;
    ccpt.conceptML = ccptML;
    dictML += ccptML;

    vector<list<Concept>::iterator>::iterator itCCPTLen = find(conceptLen[(*listIter)._len].begin()+1, conceptLen[(*listIter)._len].end(), listIter);
    assert(itCCPTLen != conceptLen[(*listIter)._len].end());
    conceptLen[(*listIter)._len].erase(itCCPTLen); 
    
    listIter = conceptList.erase(listIter);
    conceptList.insert(listIter, ccpt);
    conceptLen[len].push_back(--listIter);
    
    if(pertLegend){
        string pertLeg = "Rep : (" + to_string(ind) + "," + to_string(pos) + "," + to_string(len) + ")\n";
        file_append(logF, pertLeg);
        cout << pertLeg;
    }          
    return true;    
}

void Dictionary::_undoUpdateSwap(list<Concept>::iterator& listIter){
    assert(listIter != conceptList.end());

    vector<list<Concept>::iterator>::iterator itCCPTLen = find(conceptLen[(*listIter)._len].begin()+1, conceptLen[(*listIter)._len].end(), listIter);
    assert(itCCPTLen != conceptLen[(*listIter)._len].end());
    conceptLen[(*listIter)._len].erase(itCCPTLen); 
    
    dictML -= (*listIter).conceptML;
    listIter = conceptList.erase(listIter);
    conceptList.insert(listIter, tempCCPT);
    conceptLen[tempCCPT._len].push_back(--listIter);
    dictML += tempCCPT.conceptML;     
}

void Dictionary::populateCostMat(vector<Cost>& cmat, const string& concept){
    assert(cmat.size() == dataSet.size());
    const int nc = concept.size();
    assert(nc >= lowerBound && nc < maxLen);

    vector<vector<double>> match_m, insert_m, delete_m;
    size_t k = 0;  
    
    for(k = 0; k < dataSet.size(); k++){ 
		string sequence = dataSet[k];
		int ns = sequence.size(), jend = ns-1;

        if(nc - delta > ns) continue;                                                           //Case A: large concept compared to sequence
        else if(nc-delta > 1 && nc-delta <= ns && ns < nc+delta) jend = ns-(nc-delta);          //Case B: Maximum region ns (one alignment ns == nc-delta), upto ns-delta maximum
        else if(nc-delta < ns && ns >= nc+delta) jend = ns-(nc+delta);                          //Case C: Full region within ns  
        assert(jend >= 0 && jend < ns);

        for(int j = 0; j <= jend; j++){
            int istart = j+1, iend = ns;                                                        //Case B: end
            if(nc-delta > 1 && j+nc-delta <= ns) istart = j+nc-delta;                           //Case B: start, Case C: start
            if(j+nc+delta < ns) iend = j+nc+delta;                                              //Case C: start
            assert(iend >= istart && istart > 0 && istart <= ns && iend > 0 && iend <= ns);

            string subseq = sequence.substr(j, iend-j);
            int subLen = subseq.size();
            if(nc-delta < ns) assert(subLen >= nc-delta);
            if(nc+delta <= ns) assert(subLen <= nc+delta);
            assert(subLen > 0 && subLen <= nc+epsln);

            //Case 1: Bounded cost matrix and full history matrices 
            if(epsln >= max(subLen,nc)){
                getHistoryMatrices(concept, subseq, match_m, insert_m, delete_m);

                for(int i = istart; i <= iend; i++){
                    if((i-j) >= nc-delta && (i-j) <= nc+delta)
                        cmat[k](j,i,nc) = min({
                            match_m[i-j][nc],
                            insert_m[i-j][nc],
                            delete_m[i-j][nc]
                        }) - log(1.0/3.0) + log(regionLen); 
                    else assert(false);         
                }
            //Case 2: Bounded cost matrix and bounded history matrices  
            } else {
                getBoundedHistoryMatrices(concept, subseq, match_m, insert_m, delete_m);

                for(int i = istart; i <= iend; i++){
                    if((i-j) >= nc-delta && (i-j) <= nc+delta)
                        cmat[k](j,i,nc) = min({
                            match_m[i-j][nc],
                            insert_m[i-j][nc],
                            delete_m[i-j][nc]
                        }) - log(1.0/3.0) + log(regionLen);  
                    else assert(false); 
                }
            }
        }
    }
    //}
}

void Dictionary::getHistoryMatrices(const string& elem, const string& sequence, vector<vector<double>>& match_m, vector<vector<double>>& insert_m, vector<vector<double>>& delete_m){
    const int m = sequence.size();
    const int n = elem.size();
    
    assert(n >= lowerBound && n < maxLen);
    assert(m > 0 && m < maxLen);
    assert(match_m.size() == insert_m.size() && match_m.size() == delete_m.size());
    double inf = numeric_limits<double>::max();

    if(match_m.size() > 0){
        match_m.erase(match_m.begin(), match_m.end());
        insert_m.erase(insert_m.begin(), insert_m.end());
        delete_m.erase(delete_m.begin(), delete_m.end());
    }

    for(int p = 0; p <= m; p++){
        match_m.push_back(std::vector<double>(n+1, numeric_limits<double>::max()));
        insert_m.push_back(std::vector<double>(n+1, numeric_limits<double>::max()));
        delete_m.push_back(std::vector<double>(n+1, numeric_limits<double>::max()));
    }
    
    match_m[0][0] = 0;
    insert_m[0][0] = 0;
    delete_m[0][0] = 0;

    match_m[1][0] = inf;
    insert_m[1][0] = mmlEstimateAA(sequence[0], mmlMatrix);
    delete_m[1][0] = inf;

    for(int i=2; i<=m; i++){
        match_m[i][0] = inf;
        insert_m[i][0] = i_i + mmlEstimateAA(sequence[i-1], mmlMatrix) + insert_m[i-1][0];
        delete_m[i][0] = inf;
    }

    for(int j=1; j<=n; j++){
        match_m[0][j] = inf;
        delete_m[0][j] = d_d*(j-1);
        insert_m[0][j] = inf;
    }
 
    double scoreA, scoreB, scoreC;
    for(int i=1; i<=m; i++){
        for(int j=1; j<=n; j++){
            double sigma = conditEstimateAA(sequence[i-1], elem[j-1], condMatrix);
            double nVal = mmlEstimateAA(sequence[i-1], mmlMatrix);

            scoreA = match_m[i-1][j-1] + m_m + sigma;
            scoreB = insert_m[i-1][j-1] + m_i + sigma;
            scoreC = delete_m[i-1][j-1] + m_d + sigma;
            if(i == 1 && j == 1){
                scoreA -= m_m;
                scoreB -= m_i;
                scoreC -= m_d;
            }
            match_m[i][j] = min({scoreA, scoreB, scoreC});

            scoreA = match_m[i][j-1] + d_m;
            scoreB = insert_m[i][j-1] + d_i;
            scoreC = delete_m[i][j-1] + d_d;
            delete_m[i][j] = min({scoreA, scoreB, scoreC});

            scoreA = match_m[i-1][j]+ i_m + nVal;
            scoreB = insert_m[i-1][j] + i_i + nVal;
            scoreC = delete_m[i-1][j] + i_d + nVal;
            insert_m[i][j] = min({scoreA, scoreB, scoreC});  
       }
    }
}

void Dictionary::getBoundedHistoryMatrices(const string& elem, const string& sequence, vector<vector<double>>& match_m, vector<vector<double>>& insert_m, vector<vector<double>>& delete_m){
    const int m = sequence.size();
    const int n = elem.size();
   
    assert(n >= lowerBound && n < maxLen);
    assert(m > 0 && m < maxLen);
    assert(match_m.size() == insert_m.size() && match_m.size() == delete_m.size());
    double inf = numeric_limits<double>::max();
    int row0 = epsln, col0 = epsln;

    if(match_m.size() > 0){
        match_m.erase(match_m.begin(), match_m.end());
        insert_m.erase(insert_m.begin(), insert_m.end());
        delete_m.erase(delete_m.begin(), delete_m.end());
    }

    for(int p = 0; p <= m; p++){
        match_m.push_back(std::vector<double>(n+1, numeric_limits<double>::max()));
        insert_m.push_back(std::vector<double>(n+1, numeric_limits<double>::max()));
        delete_m.push_back(std::vector<double>(n+1, numeric_limits<double>::max()));
    }
    
    match_m[0][0] = 0;
    insert_m[0][0] = 0;
    delete_m[0][0] = 0;

    match_m[1][0] = inf;
    insert_m[1][0] = mmlEstimateAA(sequence[0], mmlMatrix);
    delete_m[1][0] = inf;

    if(m < epsln) col0 = m;
    for(int i=2; i <= col0; i++){
        match_m[i][0] = inf;
        insert_m[i][0] = i_i + mmlEstimateAA(sequence[i-1], mmlMatrix) + insert_m[i-1][0];
        delete_m[i][0] = inf;
    }

    if(n < epsln) row0 = n;
    for(int j=1; j <= row0; j++){
        match_m[0][j] = inf;
        delete_m[0][j] = d_d*(j-1);
        insert_m[0][j] = inf;
    }

    int innerStrt = 1;
    int innerEnd = n;
    assert(m <= n+epsln);
   
   double scoreA, scoreB, scoreC;
    for(int i = 1; i<=m; i++){
        if(i-epsln > 0) innerStrt = i - epsln;
        if(i+epsln <= n) innerEnd = i + epsln;

        for(int j=innerStrt; j<=innerEnd; j++){
            double sigma = conditEstimateAA(sequence[i-1], elem[j-1], condMatrix);
            double nVal = mmlEstimateAA(sequence[i-1], mmlMatrix);
            
            scoreA = match_m[i-1][j-1] + m_m + sigma;
            scoreB = insert_m[i-1][j-1] + m_i + sigma;
            scoreC = delete_m[i-1][j-1] + m_d + sigma;
            if(i == 1 && j == 1){
                scoreA -= m_m;
                scoreB -= m_i;
                scoreC -= m_d;
            }
            match_m[i][j] = min({scoreA, scoreB, scoreC});

            scoreA = match_m[i][j-1] + d_m;
            scoreB = insert_m[i][j-1] + d_i;
            scoreC = delete_m[i][j-1] + d_d;
            delete_m[i][j] = min({scoreA, scoreB, scoreC});

            scoreA = match_m[i-1][j]+ i_m + nVal;
            scoreB = insert_m[i-1][j] + i_i + nVal;
            scoreC = delete_m[i-1][j] + i_d + nVal;
            insert_m[i][j] = min({scoreA, scoreB, scoreC});  
        }
    }
}

void Dictionary::_getML(double& dictml, double& dataml){
    assert(nSize >= 1);
    if(nSize == 1) {
        dictml = bits2nits(1);
        assert(nullML > 0);
        dataml = nullML;
    } else {
        dictml = dictML + encodeWTNits(nSizeTrue);
        double encDataSize = encodeWTNits(dataSet.size());
        dataml = encDataSize;

        double totCost = 0, totEncDisect = 0, totSeql = 0; 
        int totnDisect = 0;
        size_t k;
       
        for(Concept& ccpt : conceptList){
            if(ccpt.usage.size() > 1) ccpt.usage.erase(ccpt.usage.begin()+1, ccpt.usage.end());
        }
        
        for(k = 0; k < dataSet.size(); k++){   //for each sequence in D: 
            double dissectCost = 0; 
            int nDisect = 0; 
            dissections(dataSet[k], k, dissectCost, nDisect); 
            double encnDisect = encodeWTNits(nDisect);
            double encSeqlen = encodeWTNits(dataSet[k].size());

            #pragma omp critical
            {           
                totnDisect += nDisect;                
                totEncDisect += encnDisect;
                totCost += dissectCost;
                totSeql += encSeqlen;

                dataml += encSeqlen + dissectCost + encnDisect; //length of sequence, rod cut cost, no. of dissections
            }
        }

        if(disectLegend){
            string costbrk = "----------------------- Data Cost Break Down -------------------------\n" 
            "Total dissection Cost          = " + num2str(nits2bits(totCost), 9, 10) + " bits\n" 
            "Cost of Number of dissections  = " + num2str(nits2bits(totEncDisect), 6, 10) + " bits\n"
            "Cost of stating lengths        = " + num2str(nits2bits((encDataSize+totSeql)), 6, 10) + " bits\n"
            "Total Number of dissections    = " + num2str(totnDisect, 6, 10) + "\n" 
            "Total cost                     = " + num2str(nits2bits(dataml), 9, 10) + " bits\n\n";
            file_append(logF, costbrk);
            cout << costbrk;
        }
    }
}

void Dictionary::dissections(const string& sequence, const int& seqIndx, double& dCost, int& nDisect){
    int n = sequence.size();
    assert(n > 0);
    assert(nSizeTrue > 0);
    double R[n+1];
    vector<tuple<list<Concept>::iterator,int,int>> disecParts[n+1];
    
    for(int i = 0; i <= n; i++){
        R[i] = numeric_limits<double>::max();
        disecParts[i] = vector<tuple<list<Concept>::iterator,int,int>> (1, make_tuple(conceptList.begin(),0,0));
    }

    int rgnPos = -1, rgnLen = -1, startInd = 0;
    list<Concept>::iterator ccptId = conceptList.begin();
    R[0] = 0;
    for(int i = 1; i <= n; i++){
        for (int j = 0; j < i; j++){
            double minCost = (*conceptList.begin()).costMat[seqIndx](j, i, 0) + log(regionLen);
            rgnPos = j, rgnLen = i-j, startInd = lowerBound, ccptId = conceptList.begin();

            if(rgnLen >= lowerBound){
                if(rgnLen-delta > lowerBound) startInd = rgnLen-delta; 
                for(int k = startInd; k <= rgnLen+delta; k++){
                    if(k >= conceptLen.size()) break;
                    if(conceptLen[k].size() <= 1) continue;

                    for(int p = 1; p < conceptLen[k].size(); p++){
                        assert(k == (*conceptLen[k][p])._len);
                        double cost = (*conceptLen[k][p]).costMat[seqIndx](j,i,k);
            
                        if(minCost > cost){ 
                            minCost = cost;
                            ccptId = conceptLen[k][p];
                            rgnPos = j;
                            rgnLen = i-j;
                        }   
                    }
                }
            }
            
	        minCost += (R[j] + log(nSizeTrue));
	        if(R[i] > minCost){
                R[i] = minCost;
                disecParts[i] = disecParts[j];
                disecParts[i].push_back(make_tuple(ccptId, rgnPos, rgnLen));
	        }          
        }   
    }
    nDisect = disecParts[n].size()-1;	//no of dissections
    dCost = R[n]; 			            //rod cut value (prob. of state transfer + null/dam101 ConditMat)
    assert(dCost > 0 && nDisect > 0);

    for(int p = 0; p < disecParts[n].size(); p++){
        assert(get<0>(disecParts[n][p]) != conceptList.end());
        if(get<0>(disecParts[n][p]) == conceptList.begin()) continue;
        (*get<0>(disecParts[n][p])).usage.push_back(make_tuple(seqIndx, get<1>(disecParts[n][p]), get<2>(disecParts[n][p])));
    
        if(disctBreakdown)
            cout << setw(3) << &(*get<0>(disecParts[n][p])) << " concept : (" << seqIndx << "," << get<1>(disecParts[n][p]) << "," << get<2>(disecParts[n][p]) << ")" << endl;
        
    }
   
    if(prntCost) cout << setw(3) << seqIndx << " Dissection Cost: " << setw(10) << setprecision(8) << R[n] << " - " << setw(3) << nDisect << " dissections" << endl;
}

int Dictionary::_size(){
    return nSize-1;
}

void Dictionary::_checkPoint(const string& f_name){
    string content = "";
    for(const Concept& ccpt : conceptList){
        if(ccpt._len <= 0) continue;
        content += to_string(ccpt._ind) + " " + to_string(ccpt._pos) + " " + to_string(ccpt._len) + "\n";
    }
    file_append(f_name, content);
}

void Dictionary::_loadCheckPoint(const vector<string>& content){
    assert(content.size() > 1);
    int ind = -1, pos = -1, len = -1;

    for(int j = 1; j < content.size(); j++){
        istringstream iss(content[j]);
        vector<string> legSplit((istream_iterator<string>(iss)),
                                 istream_iterator<string>());
        assert(legSplit.size() > 2);
        cout << setw(3) << j << " of " << content.size()-1 << " - ";
        ind = stoi(legSplit[0]), pos = stoi(legSplit[1]), len = stoi(legSplit[2]);
        _add(ind, pos, len);
    }
}

void Dictionary::_printUsage(vector<string>& content, const bool& wrt, const bool& dbrkdwn){
    
    string header1 = "\n----------------------- Dictionary Breakdown ------------------------";
    if(wrt) content.push_back(header1);
    cout << header1 << endl;

    for(const vector<list<Concept>::iterator> cclen : conceptLen){
        if(cclen.size() <= 1) continue;
        for(const list<Concept>::iterator ccpt : cclen){
            if(ccpt == conceptList.begin()) continue;
            if((*ccpt)._len <= 0) continue;
            stringstream is, ps, ls, usgs;
            is << setw(3) << (*ccpt)._ind;
            ps << setw(3) << (*ccpt)._pos;
            ls << setw(3) << (*ccpt)._len;
            usgs << setw(3) << (*ccpt).usage.size()-1;
            
            string ccptDesc = "# (" + is.str() + "," + ps.str() + "," + ls.str() + ") " + usgs.str() + " " + dataSet[(*ccpt)._ind].substr((*ccpt)._pos, (*ccpt)._len);
            if(wrt) content.push_back(ccptDesc);
            cout << ccptDesc << endl;
        }
    } 
    string footer = "---------------------------------------------------------------------\n";
    if(wrt) content.push_back(footer);
    cout << footer << endl << endl;

    string header3 = "\n------------------------- Usage Breakdown -------------------------";
    if(wrt) content.push_back(header3);
    if(dbrkdwn) cout << header3 << endl;

    int j = 1;
    for(const vector<list<Concept>::iterator> cclen : conceptLen){
        if(cclen.size() <= 1) continue;
        for(const list<Concept>::iterator ccpt : cclen){
            if(ccpt == conceptList.begin()) continue;
            if((*ccpt)._len <= 0) continue;
            stringstream indxs, is, ps, ls;
            indxs << setw(3) << j;
            is << setw(3) << (*ccpt)._ind;
            ps << setw(3) << (*ccpt)._pos;
            ls << setw(3) << (*ccpt)._len;
            string temps = "# " + indxs.str() + " (" + is.str() + "," + ps.str() + "," + ls.str() + ") " +
            dataSet[(*ccpt)._ind].substr((*ccpt)._pos, (*ccpt)._len) + "\n";

            for(size_t i = 1; i < (*ccpt).usage.size(); i++){
                if(get<2>((*ccpt).usage[i]) <= 0) continue;
                stringstream ius, pus, lus;
                ius << setw(3) << get<0>((*ccpt).usage[i]);
                pus << setw(3) << get<1>((*ccpt).usage[i]);
                lus << setw(3) << get<2>((*ccpt).usage[i]);
                temps += "\t@ (" + ius.str() + "," + pus.str() + "," + lus.str() + ") " + dataSet[get<0>((*ccpt).usage[i])].substr(get<1>((*ccpt).usage[i]), get<2>((*ccpt).usage[i])) + "\n";
            } 
            if(dbrkdwn) cout << temps << endl;
            if(wrt) content.push_back(temps);
            j++;
        }
   }
}
