#include "util.h"

void file_write(string& f_name, vector<string>& content){
    ofstream wfile (f_name);
    if (wfile.is_open()){
        for(auto val : content)
            wfile << val << "\n";
        wfile.close();
    } else{ 
        cerr << "Compass: I/O Error - Error writing to '"<<f_name<<"'." << endl;
        exit(1);
    }
}

void file_append(const string& f_name, const string& content){
    std::ofstream outfile;
    outfile.open(f_name, std::ios_base::app); // append instead of overwrite
    outfile << content; 
}

vector<string> file_read(const string& f_name){
    vector<string> content;
    ifstream file(f_name);
    if (file.is_open()){   //checking whether the file is open
        string str;
        while (getline(file, str)) {
            content.push_back(str);
        }
        file.close(); //close the file object.
    } else{ 
        cerr << "Compass: I/O Error - Error reading '"<<f_name<<"' file." << endl;
        exit(1);
    }
    return content;
}


int catalan(int n) {
    assert(n >= 0);
    if (n < 2) 
        return 1;

    double mult = 1;
    for (int i = 2; i <= n; i++) {
        mult *= (double)(n+i)/i;
    }
    round(mult);
    return (int)mult;
}

double encodeWTBits(int n) {
    n++;                //map from 0 onwards
    assert(n>0);        // only non-negative integers

    if (n==1) return 1;

    int cumul = 0;
    int k = 0;
    while (cumul < n) {
        cumul += catalan(k++);
    }
    return 2*(k-1)+1;
}

double encodeWTNits(int n) {
    return bits2nits(encodeWTBits(n));
}

double encodeNats(int n){
    int count, enc = 0;
    assert(n > 0);
    if(n==1) return (double) 1/log2(exp((double) 1));
    do{
        count = 0;
        while(n){
            count++;
            n = n >> 1;
        }
        enc += count; 
        n = count - 1;
    } while(n != 1);
    return (double)(enc + (double)1)/log2(exp((double) 1));
}

double encodeLogStarBits(size_t n) { 
	assert(n>=0) ;
    if (n == 0) return 0;

	double norm_const = log(2.865)/log(2); //in bits
	if (n == 1) return norm_const;

	double val = log(n)/log(2); //in bits
	double prev = val;

	while(val >= 0) {
		double t = log(prev)/log(2); //in bits
		if( t <= 0 ) break;
		val += t;
		prev = t;
	}	
	return val+norm_const;
}

double encodeLogStarNits(size_t n) { 
	assert(n>=0) ;
    if (n == 0) return 0;
    double val = encodeLogStarBits(n);
    return bits2nits(val); // in nits
}

double nits2bits(double x) {
    static double log2 = log(2) ; // log 2 to the base e
    return x/log2 ;
}

double bits2nits(double x) {
    static double log2 = log(2) ; // log 2 to the base e
    return x*log2 ;
}



string num2str(const double num, unsigned int prec, unsigned int width){
    stringstream ss;
    string ret = "";

    ss << setw(width) << setprecision(prec) << num;
    ret = ss.str();
    assert(ret.length() > 0);
    return ret;
};

//***
//***Function: ConditMat class constructors***
//usage: populate mml probability vector
//args: two arguments: file name containing mml probabilities and substitution matrix file
//***
void init(string prob_fname, string sub_fname, vector<double>& mmlMatrix, vector<double>& condMatrix){
    mmlProbVector(prob_fname, mmlMatrix);    
    condProbVector(sub_fname, condMatrix);
}

//***
//***Function: mmlProbVector method***
//Usage: normalize the mml probabilities in the file written in mml_prob2file function and return the vector containing
//mml probabilities of each amino acid in the order of codes string
//args: one arguments: file name to read mml probabilities (From mml_prob2file function)
//return: none
//***
void mmlProbVector(string& prob_fname, vector<double>& mmlMat){
    vector<double> tempVec;
    double sum = 0;
    string line;

    ifstream rfile(prob_fname);
    if (rfile.is_open()){
        while (getline (rfile,line)){
            istringstream iss(line);
            double p;
            iss >> p;
            tempVec.push_back(p);
            sum += p;   
        }
        rfile.close();
        
        assert(tempVec.size() == AA_codes.size());
        for(int i = 0; i < tempVec.size(); i++){
            int indx = AA_codes[i] - 'A';
            assert(indx >= 0 && indx < mmlMat.size());
            mmlMat[indx] = tempVec[i]/sum;
        }

    } else {
        cerr << "Compass: I/O Error - Error opening '"<<prob_fname<<"'." << endl;
        exit(1);
    }
}

//***
//***Function: condProbVector method***
//Usage: populate the blosum62 2d vector using the blosum62 matrix containing conditional probabilities
//of amino acids in the given file
//args: two arguments: file name to read substitution probabilities (BLOSUM62 matrix) and amino acid codes
//return: none
//***
void condProbVector(string& sub_fname, vector<double>& condMat){
    ifstream rfile(sub_fname);
    int k = 0;
    string line;

    if (rfile.is_open()){
        while (getline (rfile,line)){
            if(!line.empty() && isdigit(line[0])){
                istringstream iss(line);
                vector<double> tokens;
                copy(istream_iterator<double>(iss),
                    istream_iterator<double>(),
                    back_inserter(tokens));
                
                for(int i=0; i<tokens.size(); i++){
                    int row = (int)AA_codes[k] - 'A';
                    int col = (int)AA_codes[i] - 'A';
                    condMat[row * residueSize + col] = tokens[i];
                }
                k++;  
            }
        }
        rfile.close();
    } else {
        cerr << "Compass: I/O Error - Error opening '"<<sub_fname<<"'." << endl;
        exit(1);
    }
}


//***
//***Function: get_null_estimate method***
//Usage: get the negative log of mml probabilities for the null ConditMat using the probabilities in the probs_mml vector 
//populated in mmlProbVector function 
//args: two arguments: string/character to calculate null ConditMat estimate, amino acid codes
//return: null ConditMat estimate
//***
double mmlEstimate(const string& subseq, vector<double>& mmlMat){
    double est = 0;
    assert(subseq.size() > 0);
    for(int i = 0; i < subseq.size(); i++){
        int row = subseq[i] - 'A';
        assert(row >= 0 && row < mmlMat.size());
        est += -log(mmlMat[row]);
    }
    return est;
}

double mmlEstimateAA(const char& subseq, vector<double>& mmlMat){
    double est = 0;
    int row = subseq - 'A';
    assert(row >= 0 && row < mmlMat.size());
    est += -log(mmlMat[row]);
    return est;
}



//***
//***Function: get_subs_estimate method***
//Usage: get the negative log of substitution probability from blosum62 conditional probability matrix 
//populated in mmlProbVector function 
//args: three arguments: string/character to calculate substitution ConditMat estimate and the dictionary element
//and amino acid codes
//return: substitution ConditMat estimate
//***
double conditEstimate(const string& sequence, const string& element, vector<double>& condMat){
    double est = 0;
    assert(element.size() == sequence.size());
    for(int i = 0; i<element.size(); i++){
        int row = (int)sequence[i] - 'A';
        int col = (int)element[i] - 'A';
        assert(row >= 0 && row < condMat.size() && col >= 0 && col < condMat.size());
        est += -log(condMat[row * residueSize + col]);
    }
    assert(est > 0);
    return est;
}

double conditEstimateAA(const char& sequence, const char& element, vector<double>& condMat){
    double est = 0;
    int row = (int)sequence - 'A';
    int col = (int)element - 'A';
    assert(row >= 0 && row < condMat.size() && col >= 0 && col < condMat.size());
    est += -log(condMat[row * residueSize + col]);
    return est;
}



void find_non_AA_characters(string& str){
    string non_AA_codes = "BJOUXZ";
    transform(str.begin(), str.end(),str.begin(), ::toupper);
    
    for(char code : non_AA_codes){
        str.erase(remove(str.begin(), str.end(), code), str.end());
    }
}

void read_fasta_file(const string& file_name, vector<string>& data_set, const int& minlen, const int& seqLB, const int& seqUB, int& invalseq, int& ignored, int& maxReg){
    ifstream input(file_name);
    locale loc;

    if(!input.good()){
        cerr << "Compass: I/O Error --- Error in opening '"<<file_name<<"'." << endl;
        exit(1);
    }

    string line, name, content;
    while(getline( input, line ).good()){
        if(line.empty() || line[0] == '>'){ //Identifier marker of header
            //Add the sequence read from the last entry into the vector seq
            if(!name.empty()){  
                if(content.length() >= seqLB && content.length() <= seqUB){    
                    
                    //if(find(data_set.begin(), data_set.end(), content) == data_set.end()){
                        data_set.push_back(content);
                        int cntLen = content.length();
                        if(cntLen > maxReg) maxReg = cntLen;
                    /*} else {
                        invalseq += 1;
                    }*/
                } else ignored += 1;
                name.clear();
            }
            //get the name from the header
            if(!line.empty()){
                name = line.substr(1);
            }
            
            content.clear();
        } else if(!name.empty()){
            //add each line of sequence to temp variable content
            if(line.find(' ') != string::npos){ // Invalid sequence--no spaces allowed
                name.clear();
                content.clear();
                invalseq += 1;
            } else {    //ignore non amino acid characters
                find_non_AA_characters(line);
                content += line;
            }
        }
    }
    
     //Add the sequence read from the last entry into the vector seq
    if(!name.empty()){ 
        if(!line.empty()){
            transform(line.begin(), line.end(),line.begin(), ::toupper);
            content += line;
        }
        if(content.length() >= seqLB && content.length() <= seqUB){    
            //if(find(data_set.begin(), data_set.end(), content) == data_set.end()){
                data_set.push_back(content);
                int cntLen = content.length();
                if(cntLen > maxReg) maxReg = cntLen;
            //} else invalseq += 1;
           
        } else ignored += 1;
    }         
}


string readSource(const string& file_name, const bool& pathFlag, vector<string>& dataSet, const int& minlen, const int& seqLB, const int& seqUB, int& maxReg){
    int invalseq = 0, ignored = 0;

    if(pathFlag){
        vector<string> fnames = file_read(file_name);
        for(int i = 0; i < fnames.size(); i++){
            read_fasta_file(fnames[i], dataSet, minlen, seqLB, seqUB, invalseq, ignored, maxReg);
        }
    } else {
        read_fasta_file(file_name, dataSet, minlen, seqLB, seqUB, invalseq, ignored, maxReg);
    }

    maxReg *= 2;

    string dataHeader = "\n---------------- Filtering Steps: Source Collection -----------------";

    if(invalseq)
        dataHeader += "\n- No. of ignored unique sequences                 : " + to_string(invalseq);
    if(ignored)
        dataHeader += "\n- No. of sequences excluded using chain list      : " + to_string(ignored);
    if(invalseq || ignored)
        dataHeader += "\n------------------------------ Summary ------------------------------"
                    "\n- Total number of sequences ignored               : " + to_string(invalseq+ignored);
        dataHeader += "\n- Total number of sequences considered            : " + to_string(dataSet.size()) + 
                    "\n---------------------------------------------------------------------\n";
    
    cout << dataHeader;
    return dataHeader;
}